/*
 * Copyright © 2020 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of othello-core.
 *
 * Othello-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Othello-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * othello-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.othello.core.domain.impl;

import de.fhdw.gaming.core.domain.AbstractPlayer;
import de.fhdw.gaming.othello.core.domain.OthelloFieldState;
import de.fhdw.gaming.othello.core.domain.OthelloPlayer;

/**
 * Implements {@link OthelloPlayer}.
 */
final class OthelloPlayerImpl extends AbstractPlayer<OthelloPlayer> implements OthelloPlayer {

    /**
     * {@code true} if this player uses the black tokens, or {@code false} if she uses the white tokens.
     */
    private final boolean usingBlackTokens;

    /**
     * Creates an Othello player.
     *
     * @param name             The name of the player.
     * @param usingBlackTokens {@code true} if this player uses the black tokens, or {@code false} if she uses the white
     *                         tokens.
     */
    OthelloPlayerImpl(final String name, final boolean usingBlackTokens) {
        super(name);
        this.usingBlackTokens = usingBlackTokens;
    }

    /**
     * Creates an Othello player.
     *
     * @param source The {@link OthelloPlayer} to copy.
     */
    OthelloPlayerImpl(final OthelloPlayer source) {
        super(source);
        this.usingBlackTokens = source.isUsingBlackTokens();
    }

    @Override
    public String toString() {
        return String.format(
                "OthelloPlayer[name=%s, tokens=%s, state=%s, outcome=%s]",
                this.getName(),
                this.usingBlackTokens ? OthelloFieldState.BLACK.toString() : OthelloFieldState.WHITE.toString(),
                this.getState(),
                this.getOutcome());
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof OthelloPlayerImpl) {
            final OthelloPlayerImpl other = (OthelloPlayerImpl) obj;
            return super.equals(obj) && this.usingBlackTokens == other.usingBlackTokens;
        }
        return false;
    }

    @Override
    public int hashCode() {
        return super.hashCode() ^ Boolean.hashCode(this.usingBlackTokens);
    }

    @Override
    public boolean isUsingBlackTokens() {
        return this.usingBlackTokens;
    }

    @Override
    public OthelloPlayer deepCopy() {
        return new OthelloPlayerImpl(this);
    }
}
